# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SMS Data Normalization
"""

from __future__ import unicode_literals, absolute_import

import decimal
import logging


log = logging.getLogger(__name__)


class CostNormalizer(object):
    """
    Normalizer for product cost data.
    """

    def get_vendor_id(self, sms_cost):
        """
        Normalize a value for :attr:`rattail:rattail.db.model.Vendor.id`.
        """
        return sms_cost.F27

    def get_code(self, sms_cost):
        """
        Normalize a value for :attr:`rattail:rattail.db.model.ProductCost.code`.
        """
        return (sms_cost.F26 or '').strip() or None

    def get_case_size(self, sms_cost):
        """
        Normalize a value for :attr:`rattail:rattail.db.model.ProductCost.case_size`.
        """
        # Case size is a float within SMS, but integer in Rattail.
        try:
            case_size = int(sms_cost.F19)
        except ValueError:
            log.warning("case size for product {0} is not integer, but will be "
                        "coerced: {1}".format(sms_cost.F01, repr(sms_cost.F19)))
            case_size = int(decimal.Decimal(sms_cost.F19))
        return case_size or 1

    def get_case_cost(self, sms_cost):
        """
        Normalize a value for :attr:`rattail:rattail.db.model.ProductCost.case_cost`.
        """
        # For some reason these values may exist as floats in some cases, I
        # think due to FreeTDS version differences?  In any event we want
        # to ensure decimal values.
        case_cost = sms_cost.F38
        if isinstance(case_cost, float):
            case_cost = decimal.Decimal('{0:0.5f}'.format(case_cost))
        return case_cost

    def get_unit_cost(self, sms_cost):
        """
        Normalize a value for :attr:`rattail:rattail.db.model.ProductCost.unit_cost`.
        """
        # For some reason these values may exist as floats in some cases, I
        # think due to FreeTDS version differences?  In any event we want
        # to ensure decimal values.
        unit_cost = sms_cost.F1140 or 0
        if isinstance(unit_cost, float):
            unit_cost = decimal.Decimal('{0:0.5f}'.format(unit_cost))

        # If we know the case cost, calculate unit cost from that instead.
        case_cost = self.get_case_cost(sms_cost)
        if case_cost:
            unit_cost = case_cost / self.get_case_size(sms_cost)

        # Make sure our precision matches.
        if unit_cost:
            unit_cost = unit_cost.quantize(decimal.Decimal('0.00001'))
        return unit_cost
