# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importing from LOC SMS
"""

from __future__ import unicode_literals, absolute_import

import datetime
import decimal
import logging

from sqlalchemy import orm
from sqlalchemy.orm import joinedload, joinedload_all

from locsms.db import model as locsms

from rattail import enum
from rattail.db import model
from rattail.db import newimporting as importing
from rattail.gpc import GPC
from rattail.db.util import maxlen
from rattail.time import localtime, make_utc
from rattail.util import OrderedDict
from rattail.db.model.people import get_person_display_name

from rattail_locsms.db import normalize, Session as SmsSession


IGNORED_OPERATOR_NUMBERS = (1, 2, 3, 4, 5, 6, 7, 999)

log = logging.getLogger(__name__)


class LocSmsImportHandler(importing.SQLAlchemyImportHandler):
    """
    Import handler for data coming from a LOC SMS database.
    """

    def make_host_session(self):
        return SmsSession()

    def get_importers(self):
        importers = OrderedDict()
        importers['Store'] = StoreImporter
        importers['Employee'] = EmployeeImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['Customer'] = CustomerImporter
        importers['Vendor'] = VendorImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        return importers


class SmsImporter(importing.SQLAlchemyImporter):
    """
    Base class for all LOC SMS data importers.
    """

    @property
    def normalize_progress_message(self):
        return "Reading {} data from LOC SMS".format(self.model_name)

    def query(self):
        """
        Only include records whose changes have already been deployed.
        """
        q = super(SmsImporter, self).query()
        if hasattr(self.host_model_class, 'F1001'):
            q = q.filter(self.host_model_class.F1001 == 0)
        return q

    def make_utc(self, time):
        if time is None:
            return None
        return make_utc(localtime(self.config, time))

    def make_time(self, date, time):
        """
        Combine a date and time object to create a proper datetime.
        """
        if date is None:
            return None
        if time is None:
            time = datetime.time(0) # assume midnight
        else:
            time = datetime.datetime.strptime(time, '%H%M').time()
        return datetime.datetime.combine(date, time)


class StoreImporter(SmsImporter, importing.model.StoreImporter):
    """
    Importer for store data from LOC SMS.
    """
    host_model_class = locsms.Store
    key = 'id'
    supported_fields = [
        'id',
        'name',
        'phone_number',
        'fax_number',
    ]

    def normalize_source_record(self, store):
        return {
            'id': store.F1056,
            'name': store.F1531,
            'phone_number': store.F1536 or None,
            'fax_number': store.F1537 or None,
        }


class EmployeeImporter(SmsImporter, importing.model.EmployeeImporter):
    """
    Importer for employee data from LOC SMS.
    """
    host_model_class = locsms.Operator
    key = 'id'
    supported_fields = [
        'id',
        'customer_id',
        'first_name',
        'last_name',
        'full_name',
        'display_name',
        'status',
        'phone_number',
        'phone_number_2',
        'email_address',
    ]
    status_map = {
        'ACTV': enum.EMPLOYEE_STATUS_CURRENT,
        'FRMR': enum.EMPLOYEE_STATUS_FORMER,
    }

    def normalize_source_record(self, employee):
        if employee.F1126 not in IGNORED_OPERATOR_NUMBERS:
            first_name = (employee.F1143 or '').strip()
            last_name = (employee.F1144 or '').strip()
            data = {
                'id': employee.F1126,
                'customer_id': employee.F1148 or None,
                'first_name': first_name,
                'last_name': last_name,
                'full_name': get_person_display_name(first_name, last_name),
                'display_name': (employee.F1127 or '').strip(),
                'status': self.status_map.get((employee.F1552 or '').upper()),
                'phone_number': employee.F1563 or None,
                'phone_number_2': employee.F1564 or None,
                'email_address': employee.F1571 or None,
            }
            self.prioritize_2(data, 'phone_number')
            return data


class CustomerGroupImporter(SmsImporter, importing.model.CustomerGroupImporter):
    """
    Importer for customer group data from LOC SMS.
    """
    host_model_class = locsms.CustomerGroup
    key = 'id'
    supported_fields = [
        'id',
        'name',
    ]

    def normalize_source_record(self, group):
        return {
            'id': group.F1154 or None,
            'name': group.F1268,
        }


class CustomerImporter(SmsImporter, importing.model.CustomerImporter):
    """
    Importer for customer data from LOC SMS.
    """
    host_model_class = locsms.Customer
    key = 'id'
    supported_fields = [
        'id',
        'name',
        'first_name',
        'last_name',
        'phone_number',
        'phone_number_2',
        'email_address',
        'group_id',
        'group_id_2',
    ]

    def normalize_source_record(self, customer):
        data = {
            'id': customer.F1148 or None,
            'name': (customer.F1155 or '').strip(),
            'first_name': (customer.F1149 or '').strip(),
            'last_name': (customer.F1150 or '').strip(),
            'phone_number': (customer.F1172 or '').strip() or None,
            'phone_number_2': (customer.F1173 or '').strip() or None,
            'email_address': (customer.F1573 or '').strip() or None,
            'group_id': (customer.F1154 or '').strip() or None,
            'group_id_2': (customer.F1777 or '').strip() or None,
        }
        self.prioritize_2(data, 'phone_number')
        self.prioritize_2(data, 'group_id')
        return data


class VendorImporter(SmsImporter, importing.model.VendorImporter):
    """
    Importer for vendor data from LOC SMS.
    """
    host_model_class = locsms.Vendor
    key = 'id'
    supported_fields = [
        'id',
        'name',
        'special_discount',
        'phone_number',
        'phone_number_2',
        'fax_number',
        'fax_number_2',
        'contact_name',
        'contact_name_2',
        'email_address',
    ]
    maxlen_name = maxlen(model.Vendor.name)

    def normalize_source_record(self, vendor):

        # Warn about truncation until Rattail schema is addressed.
        name = vendor.F334 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("vendor name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        special_discount = vendor.F347
        if isinstance(special_discount, float):
            special_discount = decimal.Decimal(unicode(special_discount))

        data = {
            'id': vendor.F27,
            'name': name,
            'special_discount': special_discount,
            'phone_number': vendor.F341 or None,
            'phone_number_2': vendor.F1890 or None,
            'fax_number': vendor.F342 or None,
            'fax_number_2': vendor.F1891 or None,
            'contact_name': vendor.F335 or None,
            'contact_name_2': vendor.F1889 or None,
            'email_address': vendor.F2603 or None,
        }
        self.prioritize_2(data, 'phone_number')
        self.prioritize_2(data, 'fax_number')
        self.prioritize_2(data, 'contact_name')
        return data


class DepartmentImporter(SmsImporter, importing.model.DepartmentImporter):
    """
    Importer for department data from LOC SMS.
    """
    host_model_class = locsms.Department
    key = 'number'
    supported_fields = [
        'number',
        'name',
    ]
    maxlen_name = maxlen(model.Department.name)

    def normalize_source_record(self, department):

        # Warn about truncation until Rattail schema is addressed.
        name = department.F238 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("department name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'number': department.F03 or None,
            'name': name,
        }


class SubdepartmentImporter(SmsImporter, importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from LOC SMS.
    """
    host_model_class = locsms.Subdepartment
    key = 'number'
    supported_fields = [
        'number',
        'name',
        'department_number',
    ]
    maxlen_name = maxlen(model.Subdepartment.name)

    def normalize_source_record(self, subdepartment):

        # Warn about truncation until Rattail schema is addressed.
        name = subdepartment.F1022 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("subdepartment name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'number': subdepartment.F04 or None,
            'name': name,
            'department_number': subdepartment.F03 or None,
        }


class FamilyImporter(SmsImporter, importing.model.FamilyImporter):
    """
    Importer for family data from LOC SMS.
    """
    host_model_class = locsms.Family
    key = 'code'
    supported_fields = [
        'code',
        'name',
    ]
    maxlen_name = maxlen(model.Family.name)

    def normalize_source_record(self, family):

        # Warn about truncation until Rattail schema is addressed.
        name = family.F1040 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("family name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'code': family.F16 or None,
            'name': name,
        }


class ReportCodeImporter(SmsImporter, importing.model.ReportCodeImporter):
    """
    Importer for report code data from LOC SMS.
    """
    host_model_class = locsms.ReportCode
    key = 'code'
    supported_fields = [
        'code',
        'name',
    ]
    maxlen_name = maxlen(model.ReportCode.name)

    def normalize_source_record(self, reportcode):

        # Warn about truncation until Rattail schema is addressed.
        name = reportcode.F1024 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("report code name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'code': reportcode.F18 or None,
            'name': name,
        }


class ProductImporter(SmsImporter, importing.model.ProductImporter):
    """
    Importer for product data from LOC SMS.
    """
    host_model_class = locsms.Product
    key = 'upc'
    supported_fields = [
        'upc',
        'brand_name',
        'description',
        'size',
        'unit_of_measure',
        'weighed',
        'not_for_sale',
        'department_number',
        'subdepartment_number',
        'family_code',
        'report_code',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_pack_price',
        'regular_price_pack_multiple',
        'regular_price_type',
        'regular_price_level',
        'regular_price_starts',
        'regular_price_ends',
        'sale_price_price',
        'sale_price_multiple',
        'sale_price_pack_price',
        'sale_price_pack_multiple',
        'sale_price_type',
        'sale_price_level',
        'sale_price_starts',
        'sale_price_ends',
    ]
    maxlen_description = maxlen(model.Product.description)
    maxlen_size = maxlen(model.Product.size)

    def query(self):
        return self.host_session.query(locsms.Product)\
                                .options(orm.joinedload(locsms.Product.price))\
                                .options(orm.joinedload_all(
                                    locsms.Product.pos,
                                    locsms.POS.subdepartment,
                                    locsms.Subdepartment.department))

    def normalize_source_record(self, product):

        try:
            upc = GPC(product.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS product has invalid UPC: {0}".format(repr(product.F01)))
            upc = None

        weighed = False
        uom = enum.UNIT_OF_MEASURE_EACH
        pos = product.pos.get('PAL')
        if pos:
            if pos.F106 == '1':
                if pos.subdepartment and pos.subdepartment.F82 == '1':
                    weighed = True
                    uom = enum.UNIT_OF_MEASURE_POUND
            elif pos.F82 == '1':
                weighed = True
                uom = enum.UNIT_OF_MEASURE_POUND

        # Warn about truncation until Rattail schema is addressed.
        description = (product.F29 or '').strip()
        if self.maxlen_description and len(description) > self.maxlen_description:
            log.warning("product description is more than {} chars and will be truncated: {}".format(
                    self.maxlen_description, repr(description)))
            description = description[:self.maxlen_description]

        # Warn about truncation until Rattail schema is addressed.
        size = product.F22 or ''
        if self.maxlen_size and len(size) > self.maxlen_size:
            log.warning("product size is more than {} chars and will be truncated: {}".format(
                    self.maxlen_size, repr(size)))
            size = size[:self.maxlen_size]

        # Determine if sale price is currently active; if it is not then we
        # will declare None for all sale fields.
        price = product.price
        sale_starts = self.make_utc(price.sale_start if price else None)
        sale_ends = self.make_utc(price.sale_stop if price else None)
        sale_price = price.F136 if price else None
        if sale_price:
            active = False
            if sale_starts and sale_ends:
                if sale_starts <= self.now <= sale_ends:
                    active = True
            elif sale_starts:
                if sale_starts <= self.now:
                    active = True
            elif sale_ends:
                if self.now <= sale_ends:
                    active = True
            else:
                active = True
            if not active:
                sale_price = None

        # Take discount(s) into account when calculating sale prices.
        sale_pack_price = price.F139 if price else None
        if sale_price:

            discount_amount = price.F1220
            if discount_amount:
                sale_price -= discount_amount

            discount_percent = decimal.Decimal(unicode(price.F1221)) if price.F1221 else None
            if discount_percent:
                sale_price = (sale_price * (100 - discount_percent) / 100)\
                    .quantize(decimal.Decimal('0.01'))
                if sale_pack_price:
                    sale_pack_price = (sale_pack_price * (100 - discount_percent) / 100)\
                        .quantize(decimal.Decimal('0.01'))

        return {
            'upc': upc,
            'brand_name': (product.F155 or '').strip() or None,
            'description': description,
            'size': size,
            'unit_of_measure': uom,
            'weighed': weighed,
            'not_for_sale': pos.F86 == '1' if pos else False,

            'department_number': product.department.F03 if product.department else None,
            'subdepartment_number': product.subdepartment.F04 if product.subdepartment else None,
            'family_code': product.F16,
            'report_code': product.F18,

            'regular_price_price': price.F30 if price else None,
            'regular_price_multiple': (price.F31 or 1) if price else None,
            'regular_price_pack_price': price.F140 if price and price.F140 else None,
            'regular_price_pack_multiple': (price.F142 or 1) if price and price.F140 else None,
            'regular_price_type': enum.PRICE_TYPE_REGULAR if price else None,
            'regular_price_level': price.F126 if price else None,
            'regular_price_starts': self.make_utc(self.make_time(price.F35, price.F36)) if price else None,
            'regular_price_ends': self.make_utc(self.make_time(price.F129, price.F130)) if price else None,

            'sale_price_price': sale_price,
            'sale_price_multiple': (price.F135 or 1) if sale_price else None,
            'sale_price_pack_price': sale_pack_price if sale_price else None,
            'sale_price_pack_multiple': price.F143 if sale_price else None,
            'sale_price_type': enum.PRICE_TYPE_SALE if sale_price else None,
            'sale_price_level': price.F126 if sale_price else None,
            'sale_price_starts': sale_starts if sale_price else None,
            'sale_price_ends': sale_ends if sale_price else None,
        }


class ProductCodeImporter(SmsImporter, importing.model.ProductCodeImporter):
    """
    Importer for product code data from LOC SMS.
    """
    host_model_class = locsms.AlternateCode
    key = ('product_upc', 'code')
    supported_fields = [
        'product_upc',
        'primary',
        'code',
    ]

    def setup(self):
        super(ProductCodeImporter, self).setup()
        self.product_upc = None
        self.ordinal = None

    def query(self):
        return self.host_session.query(locsms.AlternateCode)\
                                .filter(locsms.AlternateCode.F1000 == 'PAL')\
                                .order_by(locsms.AlternateCode.F01,
                                          locsms.AlternateCode.F1898.desc(),
                                          locsms.AlternateCode.F154)

    def normalize_source_record(self, code):
        try:
            upc = GPC(code.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS code has invalid UPC: {0}".format(repr(code.F01)))
            upc = None

        if self.product_upc and upc == self.product_upc:
            self.ordinal += 1
        else:
            self.product_upc = upc
            self.ordinal = 1

        return {
            'product_upc': upc,
            'primary': self.ordinal == 1,
            'code': code.F154,
        }


class ProductCostImporter(SmsImporter, importing.model.ProductCostImporter):
    """
    Importer for product cost data from LOC SMS.
    """
    host_model_class = locsms.Cost
    key = ('product_upc', 'vendor_id')
    supported_fields = [
        'product_upc',
        'vendor_id',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preferred',
    ]

    def setup(self):
        super(ProductCostImporter, self).setup()
        self.product_upc = None
        self.preference = None

    def query(self):
        return self.host_session.query(locsms.Cost)\
                                .filter(locsms.Cost.F1000 == 'PAL')\
                                .filter(locsms.Cost.F1184 == 'CASE')\
                                .order_by(locsms.Cost.F01,
                                          locsms.Cost.F90.desc(),
                                          locsms.Cost.F27)

    def normalize_source_record(self, cost):
        try:
            upc = GPC(cost.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS cost has invalid UPC: {0}".format(repr(cost.F01)))
            upc = None

        normal = normalize.CostNormalizer()
        vendor_id = normal.get_vendor_id(cost)
        if upc and vendor_id:

            if self.product_upc and upc == self.product_upc:
                self.preference += 1
            else:
                self.product_upc = upc
                self.preference = 1

            return {
                'product_upc': upc,
                'vendor_id': vendor_id,
                'code': normal.get_code(cost),
                'case_size': normal.get_case_size(cost),
                'case_cost': normal.get_case_cost(cost),
                'unit_cost': normal.get_unit_cost(cost),
                'preferred': self.preference == 1,
            }
